require 'TraceDebug'
require 'FileDialog'
require 'Dialog'
require 'SourceDocument'
require 'Console'


class ActionStatusUpdater
# def enable( enabled )
end

# should probably be a module,mixin or whatever (need read doc)
class Action
  def execute()
  end

  def updateStatus( updater )
    updater.enable = enabled?()
  end

  def enabled?()
    true
  end
end


class MainFrameAction < Action
  attr_reader :mainFrame

  def initialize( mainFrame )
    @mainFrame = mainFrame
  end
end


class MainFrameDelegatedAction < MainFrameAction
  def initialize( mainFrame, executeMethod, enabledMethod =nil )
    super( mainFrame )
		@executeMethod = executeMethod
		@enabledMethod = enabledMethod
  end

  def enabled?()
    super()  &&  hasReceiver?()  &&  
        ( @enabledMethod ? receiver().send( @enabledMethod ) : true )
  end

  def execute()
    receiver().send( @executeMethod )
  end
  
  def hasReceiver?()
    false
  end
  
  def receiver()
    mainFrame().activeWindow()
  end
end


class EditorAction < MainFrameAction
  def initialize( mainFrame )
    super( mainFrame )
  end

  def editor()
    mainFrame().activeWindow()
  end

  def execute()
  end

  def enabled?()
    super()  &&  editor() &&  editor().kind_of?( SourceDocument )
  end
end


class SimpleEditorAction < MainFrameDelegatedAction
  def initialize( mainFrame, executeMethod, enabledMethod =nil )
    super( mainFrame, executeMethod, enabledMethod )
  end
  
  def editor()
    mainFrame.activeWindow()
  end

  def hasReceiver?()
    editor()  &&  editor().kind_of?( SourceDocument )  &&  receiver()
  end
end


class ConsoleOrEditorAction < SimpleEditorAction
  def initialize( mainFrame, executeMethod, enabledMethod =nil )
    super( mainFrame, executeMethod, enabledMethod )
  end
  
  def hasReceiver?()
    receiver()  &&  ( editor.kind_of?( SourceDocument )  ||  
                      editor.kind_of?( Console ) )
  end
end


class BookmarkEditorAction < SimpleEditorAction	
	def initialize( mainFrame, executeMethod, enabledMethod = nil )
		super( mainFrame, executeMethod, enabledMethod )
	end
	
	def receiver()
		editor().bookmarkManager()
	end
end


class FileNewEditorAction < MainFrameAction
  def initialize( mainFrame )
    super
  end

  def execute()
    newEditor = mainFrame.newSourceEditorWindow()
	  newEditor.title = mainFrame().nextUntitledEditorDocumentTitle()
  end
end


class FileOpenEditorAction < MainFrameAction
  def initialize( mainFrame )
    super
  end

  def execute()
    fileNames = FileDialog.openMultiSelect( [ [ 'Ruby Files (*.rb)', '*.rb' ],
																						  [ 'All Files (*.*)', '*.*' ] ],
																						'*.rb',
																						'',
																						[:mustExist] )
    if fileNames
			for fileName in fileNames
				TraceDebug.puts "FileOpen: %s" % fileName
				newEditor = mainFrame().newSourceEditorWindow()
				newEditor.loadFile( fileName ) if newEditor
			end
		end
  end
end


class FileSaveEditorAction < EditorAction
  def initialize( mainFrame, askFileName )
    super( mainFrame )
		@askFileName = askFileName
  end

  def execute()
    if !editor().fileName or @askFileName
  	  fileName = FileDialog.save( [ [ 'Ruby Files (*.rb)', '*.rb' ],
																		[ 'All Files (*.*)', '*.*' ] ],
																	'*.rb',
																	'',
																	[:overwritePrompt] )
			return nil unless fileName
			editor().fileName = fileName
		end
		TraceDebug.puts "Saving file as: "  <<  editor.fileName() 
		editor().saveFile()
  end
end


class FindEditorAction < ConsoleOrEditorAction
  def initialize( mainFrame, executeMethod, enabledMethod = nil )
    super( mainFrame, executeMethod, enabledMethod )
  end

  def receiver()
    self
  end
  
	def finder()
		mainFrame().finder()
	end
	
	def hasFinder?()
		finder() != nil
	end
			
	def findWhat()
		findWhat = Dialog.find( finder() ? finder().findWhat() : '' )
		if findWhat
			mainFrame().finder = SimpleFinder.new( findWhat )
			findNext( false )
		else
			mainFrame().finder = nil
		end
	end
	
	def findNext( skipCurrent =true )
		findPreviousOrNextOccurrence( 1, skipCurrent )
	end
	
	def findPrevious( skipCurrent =true )
		findPreviousOrNextOccurrence( -1, skipCurrent )
	end
	
	def findPreviousOrNextOccurrence( direction, skipCurrent )
		if skipCurrent
			newPos = editor().editControl().currentPos() + direction
			editor().editControl().gotoPos( newPos ) unless newPos < 0
		end
		
		editor().editControl().searchAnchor()
		searchMethod = direction < 0 ? :searchPrev : :searchNext
		if editor().editControl().send( searchMethod, 0, finder.findWhat() ) >= 0
			editor().editControl().scrollCaret()
		else
			messageSearchFailed()
		end
	end
	
	def messageSearchFailed()
		Dialog.message( 'Could not find: ' + finder().findWhat() )
	end
	
end



class SimpleFinder
	attr_accessor :findWhat
	
	def initialize( findWhat )
	  @findWhat = findWhat
	end
end


class RunUnitTestAction < EditorAction
  def initialize( mainFrame )
    super( mainFrame )
    @saveAction = SaveAllModifiedFileAction.new( mainFrame )
  end
  
  def enabled?()
    super()  &&  mainFrame().console()  &&  @saveAction.enabled?()
  end
  
  def execute()
    if !editor.dirty?()  ||  @saveAction.execute()
      console = mainFrame().console()
      slashedFileName = editor().fileName().tr( '\\', '/' ) # work around ruby ms-win32 bug
      directory = File.dirname( slashedFileName )
      scriptName = File.basename( slashedFileName )
      console.captureCommandOutput( 
                      "ruby -I #{directory} #{scriptName}", 
                      directory )
    end
  end
end


class SaveAllModifiedFileAction < MainFrameAction
  
  class SaveEditorFile < FileSaveEditorAction
    def initialize( mainFrame, editor )
      super( mainFrame, false )
      @editor = editor
    end
    
    def editor()
      @editor
    end
  end
  
  def initialize( mainFrame )
    super( mainFrame )
  end
  
  def execute()
    editors = mainFrame().editors()
    editors.each do |editor|
      next unless editor.dirty?()
      return nil unless SaveEditorFile.new( mainFrame(), editor ).execute()
    end
    true
  end
end


class MainFrame

  attr_reader :activeWindow, :console, :editors
  attr_accessor :finder

  def initialize()
    @actions = Hash.new
    @activeWindow = nil
		@untitledEditorDocumentSerial = 1
		@finder = nil
		@editors = Array.new
    @console = nil

		registerDefaultActions()
  end
	
	def activateEditControl( editControl )
    TraceDebug.puts "Activating edit control: %s" % editControl
		self.activeWindow = componentFor( editControl )
	end
  
  def inactivateEditControl( editControl )
    TraceDebug.puts "Inactivating edit control: %s" % editControl
    if editControl
      self.activeWindow = nil if self.activeWindow == componentFor( editControl )
    else
      self.activeWindow = nil
    end
  end

  def activeWindow=( window )
    oldWindow = @activeWindow ? @activeWindow.to_s() : 'NONE'
    TraceDebug.puts "Changing active window to: %s\n\twas: %s" % 
        [window, oldWindow ] if window != @activeWindow
		@activeWindow = window
  end

  def actionFor( actionId )
    @actions[actionId]
  end

  def registerAction( actionId, action )
    @actions[actionId] = action
  end

  def unregisterAction( actionId )
    @actions.delete(actionId)
  end

	# hardcoded for now...

	ID_FILE_NEW      =0xE100
	ID_FILE_OPEN     =0xE101
	ID_FILE_CLOSE    =0xE102
	ID_FILE_SAVE     =0xE103
	ID_FILE_SAVE_AS  =0xE104
	ID_EDIT_COPY     =0xE122
	ID_EDIT_CUT      =0xE123
	ID_EDIT_PASTE    =0xE125
	ID_EDIT_UNDO     =0xE12B
	ID_EDIT_REDO     =0xE12C

	ID_BOOKMARK_TOGGLE   =40000
	ID_BOOKMARK_PREVIOUS =40001
	ID_BOOKMARK_NEXT     =40002
	ID_BOOKMARK_LAST     =40003
	ID_FOLD_ALL  =40004
	ID_FIND = 40005
	ID_FIND_NEXT = 40006
	ID_FIND_PREVIOUS = 40007
  ID_RUN_UNIT_TEST = 40008
  ID_FILE_SAVE_ALL = 40009

  def registerDefaultActions()
    registerAction( ID_EDIT_COPY, ConsoleOrEditorAction.new( self, :copy ) )
    registerAction( ID_EDIT_CUT, SimpleEditorAction.new( self, :cut ) )
    registerAction( ID_EDIT_PASTE, SimpleEditorAction.new( self, :paste, :canPaste? ) )
    registerAction( ID_EDIT_UNDO, SimpleEditorAction.new( self, :undo, :canUndo? ) )
    registerAction( ID_EDIT_REDO, SimpleEditorAction.new( self, :redo, :canRedo? ) )

    registerAction( ID_FILE_OPEN, FileOpenEditorAction.new( self ) )
		registerAction( ID_FILE_NEW, FileNewEditorAction.new( self ) )
		registerAction( ID_FILE_SAVE, FileSaveEditorAction.new( self, false ) )
		registerAction( ID_FILE_SAVE_AS, FileSaveEditorAction.new( self, true ) )
    registerAction( ID_FILE_SAVE_ALL, SaveAllModifiedFileAction.new( self ) )
	
		registerAction( ID_BOOKMARK_TOGGLE, BookmarkEditorAction.new( self, :toggleBookmark ) )
		registerAction( ID_BOOKMARK_NEXT, BookmarkEditorAction.new( self, :moveToNext, :hasBookmark? ) )
		registerAction( ID_BOOKMARK_PREVIOUS, BookmarkEditorAction.new( self, :moveToPrevious, :hasBookmark? ) )
		registerAction( ID_BOOKMARK_LAST, BookmarkEditorAction.new( self, :moveToLastAdded, :hasBookmark? ) )
#		registerAction( ID_FOLD_ALL, SimpleEditorAction.new( self, :foldAll ) )
		
		registerAction( ID_FIND, FindEditorAction.new( self, :findWhat ) )
		registerAction( ID_FIND_NEXT, FindEditorAction.new( self, :findNext, :hasFinder? ) )
		registerAction( ID_FIND_PREVIOUS, FindEditorAction.new( self, :findPrevious, :hasFinder? ) )
    
    registerAction( ID_RUN_UNIT_TEST, RunUnitTestAction.new( self ) )
  end

  def nextUntitledEditorDocumentTitle
  	title = "RubySource%d" % @untitledEditorDocumentSerial
    @untitledEditorDocumentSerial += 1
		title
  end


  def newSourceEditorWindow()
    editor = SourceDocument.new( doNewEditorWindow() )
		TraceDebug.puts "New source document: %s" % editor
		@editors <<  editor
		@activeWindow = editor
		editor
  end
	
	def sourceDocumentClosed( sourceDocument )
    TraceDebug.puts "Source document closed: %s" % sourceDocument
		@editors.delete( sourceDocument )
		@activeWindow = nil unless @editors.index( @activeWindow )
	end

	def hasDirtyEditor?()
		@editors.each do |editor| 
      return true if editor.dirty?() 
    end
		false
	end
	
  def componentFor( editControl )
    component = consoleFor(editControl )
    component = sourceDocumentFor( editControl ) unless component
    component
  end  
  
  def consoleFor( editControl )
    @console if @console  &&  @console.editControl().equal?( editControl )
  end
    
	def sourceDocumentFor( editControl )
		@editors.each do |editor|
		  return editor if editor.editControl().equal?( editControl )
		end
		nil	
	end

  def attachConsoleEditControl( editControl )
    @console = Console.new( editControl )
    TraceDebug.puts "Attaching edit control to console: " + @console.to_s()
    true
  end
  
  def loadFileInNewEditor( fileName )
    sourceDocument = newSourceEditorWindow()
    sourceDocument.loadFile( fileName )
  end
  
# extension:
#  Returns an EditorControl if success, nil otherwise
#  def doNewEditorWindow()
end








