# SourceDocument

require 'scintilla/ScConstant.rb'
require 'MarkerAllocator.rb'
require 'BookmarkManager.rb'
require 'SmartIndent.rb'
require 'BraceHighlighter.rb'
require 'FoldManipulator.rb'


class SourceDocument
	include Scintilla::Constant
	
	attr_reader :editControl, :fileName, :title, :dirty, :bookmarkManager
	
	def initialize( editControl )
		@editControl = editControl
    @fileName = nil
		@title = ''
		@dirty = false
		@markerAllocator = MarkerAllocator.new( @editControl )
		setUpEditControl()
		@bookmarkManager = BookmarkManager.new( @editControl, @markerAllocator )
		setUpDirtyFlagTracker()
		setUpSmartIndent()
		setUpBraceHighlighter()
	end
	
	def to_s()
		"SourceDocument<Title=%s; FileName=%s; %s>" % [@title,@fileName,@editControl]
	end
	
	def setUpEditControl()
		setUpLexer()
		setUpLexerStyle()
		setUpMargins()
		setUpFolders()
		setUpCaret()
		setUpIndent()
		setUpRightMargin()
		setUpUndo()
    setUpWordCharacters()
    setUpBufferedDraw()
  end
	
	def setUpLexer()
    # use RUBY lexer
		@editControl.lexer = SCLEX_RUBY					# weird, it seems that 'lexer = SCLEX_RUBY' declare
																			# declare a local variable instead of calling lexer=()
								# => to investigate, using self. in the meanwhile

		# set RUBY keywords
		@editControl.setKeyWords( 0, %q[
		__LINE__ __FILE__ BEGIN END alias and begin break case class def defined? do
		else elsif end ensure false for if in module next nil not or
		redo rescue retry return self super then true undef unless until
		when while yield public protected private ' ] )		#extra space at the end needed !!!
  end
		
	def setUpLexerStyle()
		# set a flashy color for all styles, so we can spot those we forgot to set...
		for style in 0..STYLE_LASTPREDEFINED
			setStyle( style, Color::Yellow )
		end

		# setting up style colors
		setStyle( SCE_P_STRING, Color::Red )           # "..."
		setStyle( SCE_P_CHARACTER, Color::Purple )     # '...'

		setStyle( STYLE_DEFAULT, Color::Black, Color::White, 11, "Courier New" )
		setStyle( SCE_P_DEFAULT, Color::Black, Color::White, 11, "Courier New" )

		setStyle( SCE_P_WORD, Color::Blue )            # for keywords: class def module...

		setStyle( SCE_P_CLASSNAME, Color::Teal )       # name following def, class or module
		setStyle( SCE_P_DEFNAME, Color::Navy )         # name following def, class or module

		setStyle( SCE_P_STRING, Color::Red )           # "..."
		setStyle( SCE_P_CHARACTER, Color::Purple )     # '...'

		setStyle( SCE_P_OPERATOR, Color::AquaBlue )    # + - / = ...

		setStyle( SCE_P_NUMBER, Color::Maroon )        # 123.456

		setStyle( SCE_P_TRIPLEDOUBLE, Color::Green )   # =begin ... =end
		setStyle( SCE_P_COMMENTLINE, Color::Green )    # # ...

		setStyle( SCE_P_IDENTIFIER, Color::Black )     # identifiers

		setStyle( STYLE_BRACELIGHT, Color::White, Color::AquaBlue ) # brace highlight
		setStyle( STYLE_BRACEBAD, Color::White, Color::Red )	      # brace bad highlight

		setStyle( STYLE_LINENUMBER, Color::Black, Color::OffWhite ) # line number
		setStyle( STYLE_LINENUMBER, Color::Black, Color::OffWhite ) # line number
	end
		
  def setStyle( style, fore, back = Color::White, size = nil, face = nil )
    @editControl.styleSetFore( style, fore )
    @editControl.styleSetBack( style, back )
		@editControl.styleSetSize( style, size ) if size
		@editControl.styleSetFont( style, face ) if face
  end

	def setUpMargins()
		@lineNumberMargin = 0
		@symbolMargin = 1
		@folderMargin = 2
		@editControl.setMarginTypeN( @lineNumberMargin, SC_MARGIN_NUMBER )
		@editControl.setMarginWidthN( @lineNumberMargin, 38 )

		@editControl.setMarginTypeN( @symbolMargin, SC_MARGIN_SYMBOL )
		@editControl.setMarginWidthN( @symbolMargin, 16 )
		@editControl.setMarginMaskN( @symbolMargin, ~SC_MASK_FOLDERS )

		@editControl.setMarginTypeN( @folderMargin, SC_MARGIN_SYMBOL )
		@editControl.setMarginWidthN( @folderMargin, 14 )
		@editControl.setMarginMaskN( @folderMargin, SC_MASK_FOLDERS )
		@editControl.setMarginSensitiveN( @folderMargin, true )
	end
	
	def setUpFolders()
		@editControl.markerDefine( SC_MARKNUM_FOLDEROPEN, SC_MARK_MINUS )
		@editControl.markerDefine( SC_MARKNUM_FOLDER, SC_MARK_PLUS )
		@editControl.markerDefine( SC_MARKNUM_FOLDERSUB, SC_MARK_EMPTY )
		@editControl.markerDefine( SC_MARKNUM_FOLDERTAIL, SC_MARK_EMPTY )
		@editControl.markerDefine( SC_MARKNUM_FOLDEREND, SC_MARK_EMPTY )
		@editControl.markerDefine( SC_MARKNUM_FOLDEROPENMID, SC_MARK_EMPTY )
		@editControl.markerDefine( SC_MARKNUM_FOLDERMIDTAIL, SC_MARK_EMPTY )
		
		@editControl.setProperty( "fold", "1" )
		@editControl.setProperty( "fold.compact", "1" )
		@editControl.setFoldFlags( 16 )		# debug stuff, probably not needed
		
		@foldManipulator = FoldManipulator.new( self, @folderMargin )
		@editControl.addListener( :marginClick ) { | modifiers, position, margin | 
			@foldManipulator.marginFoldAction( modifiers, position, margin )
		}
	end
	
	def setUpCaret()
		@editControl.caretWidth = 2
	end
	
	def setUpRightMargin()
		@editControl.edgeColour = Color::Red
		@editControl.edgeMode = EDGE_LINE
		@editControl.edgeColumn = 80
	end
	
	def setUpUndo()
    @editControl.undoCollection = true
		@editControl.emptyUndoBuffer()
		@editControl.setSavePoint()
	end

  def setUpIndent()
    @editControl.tabWidth = 2
    @editControl.indent = @editControl.tabWidth()
    @editControl.backSpaceUnIndents = true
		@editControl.useTabs = false		# indent using spaces
	end
	
	def setUpDirtyFlagTracker()
		@editControl.addListener( :savePointReached ) { || self.dirty = false }
		@editControl.addListener( :savePointLeft ) { || self.dirty = true }
	end

	def setUpSmartIndent()
		@smartIndent = SmartIndent.new( self )
		@editControl.addListener( :charAdded ) { | ch |
		  @smartIndent.smartIndent() if ch == effectiveEndOfLineCharacter()
		}
	end
	
	def setUpBraceHighlighter()
		@braceHighlighter = BraceHighlighter.new( self )
		@editControl.addListener( :updateUI ) { ||
			@braceHighlighter.highlightBraces()
		}
	end
  
  def setUpWordCharacters()
    characters = '@_?!$'
    (0..255).each do |code|
      char = code.chr()
      if ('a'..'z') === char  || 
         ('A'..'Z') === char  ||  
         ('0'..'9') === char  ||
         code >= 128
        characters += char
      end
    end
    @editControl.wordChars = characters
  end
  
  def setUpBufferedDraw()
    @editControl.bufferedDraw = true
  end
		
  def effectiveEndOfLineCharacter()
    return "\n" if @editControl.eOLMode() == Scintilla::Constant::SC_EOL_CRLF || 
			             @editControl.eOLMode() == Scintilla::Constant::SC_EOL_LF
		return "\r"
  end		


  # load a new file in the edit control
  def loadFile( fileName )
    self.fileName = File.expand_path( fileName )
    @editControl.clearAll()
		text = ''
    File.open( fileName, 'r' ) do |file|
      file.each_line do |line|
				text << line
			end
		end
		@editControl.addText( text )
		@editControl.documentStart()
		@editControl.emptyUndoBuffer()
		@editControl.setSavePoint()
    true
  end

  def saveFile()
    File.open( fileName(), 'w' ) do |file|
			for lineNumber in 0...@editControl.lineCount()
				line = @editControl.getLine( lineNumber )
				file.puts line.gsub( /\r|\n/, '' )
			end
		end
		@editControl.setSavePoint()
		true
  end

# extension:
# def doTitle=( title )

  def title=( title )
    @title = title
		updateTitle()
  end
	
	def updateTitle()
		dirty = dirty?() ? '* ' : ''
		if @fileName 
			baseName = File.basename( @fileName.gsub( /\\/, '/') )
			path = File.dirname( @fileName.gsub( /\\/, '/') )
			@title = baseName + ' -- '  + path
		end
		doSetTitle( dirty + title() )
	end

  def fileName=( fileName )
    @fileName = fileName
		updateTitle()
  end
	
	def dirty=( isDirty )
		if @dirty != isDirty
		  @dirty = isDirty
			updateTitle()
		end
	end
	
	def dirty?
		@dirty
	end
	
	# copy, paste cut... May want to introduce helper delegate @editControl, :canPaste?, :paste...
	def canPaste?()
		@editControl.canPaste?()
	end
	
	def paste()
		@editControl.paste()
	end
	
	def copy()
		@editControl.copy()
	end
	
	def cut()
		@editControl.cut()
	end
	
	def canUndo?()
		@editControl.canUndo?()
	end
	
	def undo()
		@editControl.undo()
	end
	
	def canRedo?()
		@editControl.canRedo?()
	end
	
	def redo()
		@editControl.redo()
	end
		
end











